from enum import Enum, auto


class State(Enum):
    """Estados posibles del cliente de WhatsApp Web"""

    LOGGED_IN = auto()  # Usuario autenticado y listo para usar
    LOADING = auto()  # Cargando interfaz o datos
    QR_AUTH = auto()  # Mostrando código QR para autenticación
    AUTH = auto()  # Pantalla inicial de autenticación
    CONNECTING = auto()  # Conectando con WhatsApp Web
    ERROR = auto()  # Error en la conexión o autenticación

    def __str__(self):
        return self.name.lower()



import re
import os
from pathlib import Path
from typing import Optional, Dict, List, Any, Union
import asyncio

from playwright.async_api import(
    TimeoutError as PlaywrightTimeoutError,
    Error as PlaywrightError,
)

from .constants import locator as loc
from .object.message import Message, FileMessage


class ChatManager:
    def __init__(self, client):
        self.client = client
        self._page = client._page
        self.wa_elements = client.wa_elements

    async def _check_unread_chats(self):
        """Verifica y devuelve los chats no leídos"""
        unread_chats = []
        try:
            unread_button = await self._page.query_selector(loc.UNREAD_CHATS_BUTTON)
            if unread_button:
                await unread_button.click()
                await asyncio.sleep(self.client.unread_messages_sleep)

                chat_list = await self._page.query_selector_all(loc.UNREAD_CHAT_DIV)
                if chat_list and len(chat_list) > 0:
                    chats = await chat_list[0].query_selector_all(loc.SEARCH_ITEM)
                    for chat in chats:
                        chat_result = await self._parse_search_result(chat, "CHATS")
                        if chat_result:
                            unread_chats.append(chat_result)

            # Volver a la vista de todos los chats
            all_button = await self._page.query_selector(loc.ALL_CHATS_BUTTON)
            if all_button:
                await all_button.click()

        except Exception as e:
            await self.client.emit("on_warning", f"Error al verificar chats no leídos: {e}")

        return unread_chats

    async def _parse_search_result(
        self, element, result_type: str = "CHATS"
    ) -> Optional[Dict[str, Any]]:
        try:
            components = await element.query_selector_all(
                "xpath=.//div[@role='gridcell' and @aria-colindex='2']/parent::div/div"
            )
            count = len(components)

            unread_el = await element.query_selector(
                f"xpath={loc.SEARCH_ITEM_UNREAD_MESSAGES}"
            )
            unread_count = await unread_el.inner_text() if unread_el else "0"
            mic_span = await components[1].query_selector('xpath=.//span[@data-icon="mic"]')
            
            if count == 3:
                span_title_0 = await components[0].query_selector(
                    f"xpath={loc.SPAN_TITLE}"
                )
                group_title = (
                    await span_title_0.get_attribute("title") if span_title_0 else ""
                )

                datetime_children = await components[0].query_selector_all("xpath=./*")
                datetime_text = (
                    await datetime_children[1].text_content()
                    if len(datetime_children) > 1
                    else ""
                )

                span_title_1 = await components[1].query_selector(
                    f"xpath={loc.SPAN_TITLE}"
                )
                title = (
                    await span_title_1.get_attribute("title") if span_title_1 else ""
                )

                info_text = (await components[2].text_content()) or ""
                info_text = info_text.replace("\n", "")

                if "loading" in info_text or "status-" in info_text or "typing" in info_text:
                    return None

                return {
                    "type": result_type,
                    "group": group_title,
                    "name": title,
                    "last_activity": datetime_text,
                    "last_message": info_text,
                    "last_message_type": "audio" if mic_span else "text",
                    "unread_count": unread_count,
                    "element": element,
                }

            elif count == 2:
                span_title_0 = await components[0].query_selector(
                    f"xpath={loc.SPAN_TITLE}"
                )
                title = (
                    await span_title_0.get_attribute("title") if span_title_0 else ""
                )

                datetime_children = await components[0].query_selector_all("xpath=./*")
                datetime_text = (
                    await datetime_children[1].text_content()
                    if len(datetime_children) > 1
                    else ""
                )

                info_children = await components[1].query_selector_all("xpath=./*")
                info_text = (
                    await info_children[0].text_content()
                    if len(info_children) > 0
                    else ""
                ) or ""
                info_text = info_text.replace("\n", "")
                if "loading" in info_text or "status-" in info_text or "typing" in info_text:
                    return None

                return {
                    "type": result_type,
                    "name": title,
                    "last_activity": datetime_text,
                    "last_message": info_text,
                    "last_message_type": "audio" if mic_span else "text",
                    "unread_count": unread_count,
                    "element": element,
                    "group": None,
                }

            return None

        except Exception as e:
            print(f"Error parsing result: {e}")
            return None

    async def close(self):
        """Cierra el chat o la vista actual presionando Escape."""
        if self._page:
            try:
                await self._page.keyboard.press("Escape")
            except Exception as e:
                await self.client.emit(
                    "on_warning", f"Error trying to close chat with Escape: {e}"
                )

    async def open(
        self, chat_name: str, timeout: int = 10000, force_open: bool = False
    ) -> bool:
        return await self.wa_elements.open(chat_name, timeout, force_open)

    async def search_conversations(
        self, query: str, close=True
    ) -> List[Dict[str, Any]]:
        """Busca conversaciones por término"""
        if not await self.client.wait_until_logged_in():
            return []
        try:
            return await self.wa_elements.search_chats(query, close)
        except Exception as e:
            await self.client.emit("on_error", f"Search error: {e}")
            return []

    async def collect_messages(self) -> List[Union[Message, FileMessage]]:
        """
        Recorre todos los contenedores de mensaje (message-in/message-out) actualmente visibles
        y devuelve una lista de instancias Message o FileMessage.
        """
        resultados: List[Union[Message, FileMessage]] = []
        msg_elements = await self._page.query_selector_all(
            'div[class*="message-in"], div[class*="message-out"]'
        )

        for elem in msg_elements:
            file_msg = await FileMessage.from_element(elem)
            if file_msg:
                resultados.append(file_msg)
                continue

            simple_msg = await Message.from_element(elem)
            if simple_msg:
                resultados.append(simple_msg)

        return resultados

    async def download_all_files(self, carpeta: Optional[str] = None) -> List[Path]:
        """
        Llama a collect_messages(), filtra FileMessage y descarga cada uno.
        Devuelve lista de Path donde se guardaron.
        """
        if not await self.client.wait_until_logged_in():
            return []

        if carpeta:
            downloads_dir = Path(carpeta)
        else:
            downloads_dir = Path.home() / "Downloads" / "WhatsAppFiles"

        archivos_guardados: List[Path] = []
        mensajes = await self.collect_messages()
        for m in mensajes:
            if isinstance(m, FileMessage):
                ruta = await m.download(self._page, downloads_dir)
                if ruta:
                    archivos_guardados.append(ruta)
        return archivos_guardados

    async def download_file_by_index(
        self, index: int, carpeta: Optional[str] = None
    ) -> Optional[Path]:
        """
        Descarga sólo el FileMessage en la posición `index` de la lista devuelta
        por collect_messages() filtrando por FileMessage.
        """
        if not await self.client.wait_until_logged_in():
            return None

        if carpeta:
            downloads_dir = Path(carpeta)
        else:
            downloads_dir = Path.home() / "Downloads" / "WhatsAppFiles"

        mensajes = await self.collect_messages()
        archivos = [m for m in mensajes if isinstance(m, FileMessage)]
        if index < 0 or index >= len(archivos):
            return None

        return await archivos[index].download(self._page, downloads_dir)

    async def send_message(
        self, chat_query: str, message: str, force_open=True
    ) -> bool:
        """Envía un mensaje a un chat"""
        if not await self.client.wait_until_logged_in():
            return False

        try:
            if force_open:
                await self.open(chat_query)
            await self._page.wait_for_selector(loc.CHAT_INPUT_BOX, timeout=10000)
            input_box = await self._page.wait_for_selector(
                loc.CHAT_INPUT_BOX, timeout=10000
            )
            if not input_box:
                await self.client.emit(
                    "on_error",
                    "No se encontró el cuadro de texto para enviar el mensaje",
                )
                return False

            await input_box.click()
            await input_box.fill(message)
            await self._page.keyboard.press("Enter")
            return True

        except Exception as e:
            await self.client.emit("on_error", f"Error al enviar el mensaje: {e}")
            return False
        finally:
            await self.close()

    async def send_file(self, chat_name, path):
        """Envía un archivo a un chat especificado en WhatsApp Web usando Playwright"""
        try:
            if not os.path.isfile(path):
                msg = f"El archivo no existe: {path}"
                await self.client.emit("on_error", msg)
                return False

            if not await self.client.wait_until_logged_in():
                msg = "No se pudo iniciar sesión"
                await self.client.emit("on_error", msg)
                return False

            if not await self.open(chat_name):
                msg = f"No se pudo abrir el chat: {chat_name}"
                await self.client.emit("on_error", msg)
                return False

            await self._page.wait_for_selector(loc.CHAT_INPUT_BOX, timeout=10000)

            attach_btn = await self._page.wait_for_selector(
                loc.ATTACH_BUTTON, timeout=5000
            )
            await attach_btn.click()

            input_files = await self._page.query_selector_all(loc.FILE_INPUT)
            if not input_files:
                msg = "No se encontró input[type='file']"
                await self.client.emit("on_error", msg)
                return False

            await input_files[0].set_input_files(path)
            await self.client.asyncio.sleep(1)

            send_btn = await self._page.wait_for_selector(
                loc.SEND_BUTTON, timeout=10000
            )
            await send_btn.click()

            return True

        except Exception as e:
            msg = f"Error inesperado en send_file: {str(e)}"
            await self.client.emit("on_error", msg)
            await self._page.screenshot(path="debug_send_file/error_unexpected.png")
            return False
        finally:
            await self.close()

    async def new_group(self, group_name: str, members: list[str]):
        return await self.wa_elements.new_group(group_name, members)

    async def add_members_to_group(self, group_name: str, members: list[str]) -> bool:
        """
        Abre un grupo y le añade nuevos miembros.
        """
        try:
            # 1. Abrir el chat del grupo
            if not await self.open(group_name):
                await self.client.emit("on_error", f"No se pudo abrir el grupo '{group_name}'")
                return False

            # 2. Llamar al método de bajo nivel para agregar miembros
            success = await self.wa_elements.add_members_to_group(group_name, members)
            return success

        except Exception as e:
            await self.client.emit("on_error", f"Error al añadir miembros al grupo '{group_name}': {e}")
            return False



"""
Base classes for creating custom filters.
"""
from abc import ABC, abstractmethod
from typing import Any, Callable


class Filter(ABC):
    """
    Abstract base class for all filters.
    """

    @abstractmethod
    def test(self, value: Any) -> bool:
        """
        This method should be overridden in subclasses.
        It takes a value and returns True if the value passes the filter, False otherwise.
        """
        raise NotImplementedError

    def __call__(self, value: Any) -> bool:
        return self.test(value)


class CustomFilter(Filter):
    """
    A filter that uses a custom function to test values.
    """

    def __init__(self, func: Callable[[Any], bool]):
        self.func = func

    def test(self, value: Any) -> bool:
        return self.func(value)



import pytest
from unittest.mock import AsyncMock, MagicMock
from src.whatsplay.chat_manager import ChatManager

# Mock de loc (locator) para evitar dependencias reales
class MockLocator:
    SEARCH_ITEM_UNREAD_MESSAGES = ".//span[contains(@aria-label, 'unread message')]"
    SPAN_TITLE = ".//span[@title]"

# Mock de un cliente mínimo para ChatManager
class MockClient:
    def __init__(self):
        self._page = AsyncMock()
        self.wa_elements = MagicMock()
        self.unread_messages_sleep = 1

    async def emit(self, event, *args, **kwargs):
        pass

    async def wait_until_logged_in(self):
        return True

@pytest.fixture
def mock_chat_manager():
    client = MockClient()
    manager = ChatManager(client)
    manager.loc = MockLocator() # Inyectar el mock de loc
    return manager

@pytest.mark.asyncio
async def test_parse_search_result_individual_chat(mock_chat_manager):
    # Simular un elemento de chat individual (2 componentes)
    mock_element = AsyncMock()
    
    # Mock de components[0] (título y datetime)
    mock_component0 = AsyncMock()
    mock_span_title0 = AsyncMock()
    mock_span_title0.get_attribute.return_value = "Contacto Individual"
    mock_component0.query_selector.return_value = mock_span_title0
    
    mock_datetime_child = AsyncMock()
    mock_datetime_child.text_content.return_value = "10:30 AM"
    mock_component0.query_selector_all.return_value = [MagicMock(), mock_datetime_child]

    # Mock de components[1] (info_text)
    mock_component1 = AsyncMock()
    mock_info_child = AsyncMock()
    mock_info_child.text_content.return_value = "Hola, ¿cómo estás?"
    mock_component1.query_selector_all.return_value = [mock_info_child]

    mock_element.query_selector_all.return_value = [mock_component0, mock_component1]

    # Mock de unread_el
    mock_unread_el = AsyncMock()
    mock_unread_el.inner_text.return_value = "5"
    mock_element.query_selector.return_value = mock_unread_el

    result = await mock_chat_manager._parse_search_result(mock_element)

    assert result is not None
    assert result["type"] == "CHATS"
    assert result["name"] == "Contacto Individual"
    assert result["last_activity"] == "10:30 AM"
    assert result["last_message"] == "Hola, ¿cómo estás?"
    assert result["unread_count"] == "5"
    assert result["group"] is None

@pytest.mark.asyncio
async def test_parse_search_result_group_chat(mock_chat_manager):
    # Simular un elemento de chat de grupo (3 componentes)
    mock_element = AsyncMock()

    # Mock de components[0] (group_title y datetime)
    mock_component0 = AsyncMock()
    mock_span_title0 = AsyncMock()
    mock_span_title0.get_attribute.return_value = "Nombre del Grupo"
    mock_component0.query_selector.return_value = mock_span_title0

    mock_datetime_child = AsyncMock()
    mock_datetime_child.text_content.return_value = "Ayer"
    mock_component0.query_selector_all.return_value = [MagicMock(), mock_datetime_child]

    # Mock de components[1] (title - que es el mismo que el grupo en este caso)
    mock_component1 = AsyncMock()
    mock_span_title1 = AsyncMock()
    mock_span_title1.get_attribute.return_value = "Nombre del Grupo"
    mock_component1.query_selector.return_value = mock_span_title1

    # Mock de components[2] (info_text con remitente)
    mock_component2 = AsyncMock()
    mock_component2.text_content.return_value = "Juan Pérez: Mensaje del grupo"

    mock_element.query_selector_all.return_value = [mock_component0, mock_component1, mock_component2]

    # Mock de unread_el
    mock_unread_el = AsyncMock()
    mock_unread_el.inner_text.return_value = "12"
    mock_element.query_selector.return_value = mock_unread_el

    result = await mock_chat_manager._parse_search_result(mock_element)

    assert result is not None
    assert result["type"] == "CHATS"
    assert result["group"] == "Nombre del Grupo"
    assert result["name"] == "Nombre del Grupo"
    assert result["last_activity"] == "Ayer"
    assert result["last_message"] == "Juan Pérez: Mensaje del grupo"
    assert result["unread_count"] == "12"

@pytest.mark.asyncio
async def test_parse_search_result_group_chat_with_different_sender(mock_chat_manager):
    # Simular un elemento de chat de grupo (3 componentes) donde el remitente es diferente al título del grupo
    mock_element = AsyncMock()

    # Mock de components[0] (group_title y datetime)
    mock_component0 = AsyncMock()
    mock_span_title0 = AsyncMock()
    mock_span_title0.get_attribute.return_value = "Mi Grupo de Amigos"
    mock_component0.query_selector.return_value = mock_span_title0

    mock_datetime_child = AsyncMock()
    mock_datetime_child.text_content.return_value = "Hoy"
    mock_component0.query_selector_all.return_value = [MagicMock(), mock_datetime_child]

    # Mock de components[1] (title - que es el mismo que el grupo en este caso)
    mock_component1 = AsyncMock()
    mock_span_title1 = AsyncMock()
    mock_span_title1.get_attribute.return_value = "Mi Grupo de Amigos"
    mock_component1.query_selector.return_value = mock_span_title1

    # Mock de components[2] (info_text con remitente diferente al título del grupo)
    mock_component2 = AsyncMock()
    mock_component2.text_content.return_value = "Ana: Hola a todos!"

    mock_element.query_selector_all.return_value = [mock_component0, mock_component1, mock_component2]

    # Mock de unread_el
    mock_unread_el = AsyncMock()
    mock_unread_el.inner_text.return_value = "3"
    mock_element.query_selector.return_value = mock_unread_el

    result = await mock_chat_manager._parse_search_result(mock_element)

    assert result is not None
    assert result["type"] == "CHATS"
    assert result["group"] == "Mi Grupo de Amigos" # Sigue siendo un grupo
    assert result["name"] == "Mi Grupo de Amigos"
    assert result["last_activity"] == "Hoy"
    assert result["last_message"] == "Ana: Hola a todos!"
    assert result["unread_count"] == "3"

@pytest.mark.asyncio
async def test_parse_search_result_individual_chat_with_colon_in_message(mock_chat_manager):
    # Simular un chat individual donde el mensaje contiene dos puntos, pero no es un grupo
    mock_element = AsyncMock()
    
    # Mock de components[0] (título y datetime)
    mock_component0 = AsyncMock()
    mock_span_title0 = AsyncMock()
    mock_span_title0.get_attribute.return_value = "Contacto Individual"
    mock_component0.query_selector.return_value = mock_span_title0
    
    mock_datetime_child = AsyncMock()
    mock_datetime_child.text_content.return_value = "11:00 AM"
    mock_component0.query_selector_all.return_value = [MagicMock(), mock_datetime_child]

    # Mock de components[1] (info_text con dos puntos, pero sin remitente)
    mock_component1 = AsyncMock()
    mock_info_child = AsyncMock()
    mock_info_child.text_content.return_value = "Mensaje: con dos puntos"
    mock_component1.query_selector_all.return_value = [mock_info_child]

    mock_element.query_selector_all.return_value = [mock_component0, mock_component1]

    # Mock de unread_el
    mock_unread_el = AsyncMock()
    mock_unread_el.inner_text.return_value = "1"
    mock_element.query_selector.return_value = mock_unread_el

    result = await mock_chat_manager._parse_search_result(mock_element)

    assert result is not None
    assert result["type"] == "CHATS"
    assert result["name"] == "Contacto Individual"
    assert result["last_activity"] == "11:00 AM"
    assert result["last_message"] == "Mensaje: con dos puntos"
    assert result["unread_count"] == "1"
    assert result["group"] is None



import http.server
import socketserver
import threading
import base64
import time

# Global variable to hold the QR code and the HTTP server instance
current_qr_base64 = None
httpd = None
last_request_time = time.time()

def update_qr_code(qr_image_bytes):
    """Updates the global QR code variable."""
    global current_qr_base64
    current_qr_base64 = base64.b64encode(qr_image_bytes).decode('utf-8')

def show_qr_window(qr_image_bytes):
    """
    Muestra un QR en una página web simple con actualización en tiempo real.
    """
    update_qr_code(qr_image_bytes)

    html_content = f"""
    <html>
        <head>
            <title>WhatsApp QR Code</title>
            <style>
                body {{
                    display: flex;
                    justify-content: center;
                    align-items: center;
                    height: 100vh;
                    margin: 0;
                    background-color: #f0f0f0;
                }}
                img {{
                    max-width: 100%;
                    max-height: 100%;
                }}
            </style>
            <script>
                function refreshQR() {{
                    fetch('/qr')
                        .then(response => response.text())
                        .then(data => {{
                            document.getElementById('qr-img').src = "data:image/png;base64," + data;
                        }});
                }}
                setInterval(refreshQR, 1000);
            </script>
        </head>
        <body>
            <img id="qr-img" src="data:image/png;base64,{current_qr_base64}" alt="QR Code">
        </body>
    </html>
    """

    class Handler(http.server.SimpleHTTPRequestHandler):
        def do_GET(self):
            global last_request_time
            last_request_time = time.time()
            if self.path == '/':
                self.send_response(200)
                self.send_header('Content-type', 'text/html')
                self.end_headers()
                self.wfile.write(html_content.encode('utf-8'))
            elif self.path == '/qr':
                self.send_response(200)
                self.send_header('Content-type', 'text/plain')
                self.end_headers()
                self.wfile.write(current_qr_base64.encode('utf-8'))
            else:
                super().do_GET()

    PORT = 8000
    
    def start_server():
        global httpd
        # Bind to 0.0.0.0 to make it accessible from the public IP
        with socketserver.TCPServer(("0.0.0.0", PORT), Handler) as httpd_server:
            httpd = httpd_server
            print(f"Servidor iniciado en http://0.0.0.0:{PORT}")
            httpd.serve_forever()

    server_thread = threading.Thread(target=start_server)
    server_thread.daemon = True
    server_thread.start()

def close_qr_window():
    global httpd
    if httpd:
        httpd.shutdown()


"""
Utilities for interacting with WhatsApp Web elements
"""

import asyncio
from typing import Optional, List, Dict, Any
from playwright.async_api import (
    Page,
    ElementHandle,
    TimeoutError as PlaywrightTimeoutError,
)

from .constants import locator as loc
from .constants.states import State
from .filters import MessageFilter


class WhatsAppElements:
    """Helper class for interacting with WhatsApp Web elements"""

    def __init__(self, page: Page):
        self.page = page

    async def get_state(self) -> Optional[State]:
        """
        Determina el estado actual de WhatsApp Web basado en los elementos visibles
        """
        try:
            # Checkear en orden de prioridad
            if await self.page.locator(loc.LOGGED_IN).is_visible():
                print("LOGGED_IN")
                return State.LOGGED_IN
            elif await self.page.locator(loc.LOADING).is_visible():
                print("LOADING")
                return State.LOADING
            elif await self.page.locator(loc.QR_CODE).is_visible():
                print("QR_AUTH")
                return State.QR_AUTH
            elif await self.page.locator(loc.AUTH).is_visible():
                print("AUTH")
                return State.AUTH
            elif await self.page.locator(loc.LOADING_CHATS).is_visible():
                print("LOADING_CHATS")
                return State.LOADING
            return None
        except Exception:
            return None

    async def wait_for_selector(
        self, selector: str, timeout: int = 5000, state: str = "visible"
    ) -> Optional[ElementHandle]:
        """
        Espera por un elemento y lo retorna cuando está disponible
        """
        try:
            element = await self.page.wait_for_selector(
                selector, timeout=timeout, state=state
            )
            return element
        except PlaywrightTimeoutError:
            return None

    async def click_search_button(self) -> bool:
        """Intenta hacer click en el botón de búsqueda usando múltiples estrategias"""
        try:
            # Intentar con cada selector del botón de búsqueda
            for selector in loc.SEARCH_BUTTON:
                try:
                    element = await self.page.wait_for_selector(
                        selector, timeout=1000, state="visible"
                    )
                    if element:
                        await element.click()
                        if await self.verify_search_active():
                            return True
                except Exception:
                    continue

            # Si no funcionó el clic directo, intentar con atajos de teclado
            shortcuts = ["Control+/", "Control+f", "/", "Slash"]
            for shortcut in shortcuts:
                try:
                    await self.page.keyboard.press("Escape")  # Limpiar estado actual
                    await self.page.keyboard.press(shortcut)
                    if await self.verify_search_active():
                        return True
                except Exception:
                    continue

            return False

        except Exception as e:
            print(f"Error clicking search button: {e}")
            return False

    async def verify_search_active(self) -> bool:
        """Verifica si la búsqueda está activa usando múltiples indicadores"""
        try:
            # Verificar si el botón de cancelar búsqueda está visible
            cancel_button = await self.wait_for_selector(
                loc.CANCEL_SEARCH, timeout=1000
            )
            if cancel_button:
                return True

            # Verificar si algún campo de búsqueda está visible
            for selector in loc.SEARCH_TEXT_BOX:
                try:
                    element = await self.page.wait_for_selector(
                        selector, timeout=1000, state="visible"
                    )
                    if element:
                        return True
                except Exception:
                    continue

            return False
        except Exception:
            return False

    async def get_qr_code(self) -> Optional[bytes]:
        """
        Obtiene la imagen del código QR si está disponible
        """
        try:
            qr_element = await self.wait_for_selector(loc.QR_CODE)
            if qr_element:
                return await qr_element.screenshot()
            return None
        except Exception:
            return None

    async def search_chats(self, query: str, close=True) -> List[Dict[str, Any]]:
        """Busca chats usando un término y retorna los resultados"""
        results = []
        try:
            # Activar búsqueda
            if not await self.click_search_button():
                return results

            # Buscar campo de texto y escribir consulta
            search_box = None
            for selector in loc.SEARCH_TEXT_BOX:
                try:
                    search_box = await self.wait_for_selector(selector, timeout=2000)
                    if search_box:
                        break
                except Exception:
                    continue

            if not search_box:
                return results

            # Escribir consulta con reintento
            max_attempts = 3
            for attempt in range(max_attempts):
                try:
                    await search_box.click()
                    await search_box.fill("")
                    await search_box.type(query, delay=100)
                    break
                except Exception as e:
                    if attempt == max_attempts - 1:
                        return results

            # Esperar resultados
            results_container = await self.wait_for_selector(
                loc.SEARCH_RESULT, timeout=5000
            )
            if not results_container:
                print("No search results found")
                return results

            # Obtener y procesar resultados
            items = await self.page.locator(loc.SEARCH_ITEM).all()
            for item in items:
                text = await item.inner_text()
                if text:
                    formatted = MessageFilter.filter_search_result(text)
                    results.append(formatted)

        except Exception as e:
            print(f"Error searching chats: {e}")
        finally:
            # Cerrar búsqueda
            try:
                if close:
                    await self.page.keyboard.press("Escape")
            except:
                pass

        return results

    async def open(
        self, chat_name: str, timeout: int = 10000, force_open: bool = False
    ) -> bool:
        """
        Abre un chat por su nombre visible. Si no está en el DOM, lo busca y lo abre.
        """
        es_numero = bool(re.fullmatch(r"\+?\d+", chat_name))

        if es_numero or force_open:
            numero = chat_name.lstrip("+")
            url = f"https://web.whatsapp.com/send?phone={numero}"
            await self.page.goto(url)

        span_xpath = f"//span[contains(@title, {repr(chat_name)})]"

        try:
            chat_element = await self.page.query_selector(f"xpath={span_xpath}")
            if chat_element:
                await chat_element.click()
                print(f"✅ Chat '{chat_name}' abierto directamente.")
            else:
                print(f"🔍 Chat '{chat_name}' no visible, usando buscador...")
                for btn in loc.SEARCH_BUTTON:
                    btns = await self.page.query_selector_all(f"xpath={btn}")
                    if btns:
                        await btns[0].click()
                        break
                else:
                    raise Exception("❌ Botón de búsqueda no encontrado")

                for input_xpath in loc.SEARCH_TEXT_BOX:
                    inputs = await self.page.query_selector_all(f"xpath={input_xpath}")
                    if inputs:
                        await inputs[0].fill(chat_name)
                        break
                else:
                    raise Exception("❌ Input de búsqueda no encontrado")

                await self.page.wait_for_selector(loc.SEARCH_ITEM, timeout=timeout)
                await self.page.keyboard.press("ArrowDown")
                await self.page.keyboard.press("Enter")
                print(f"✅ Chat '{chat_name}' abierto desde buscador.")

            await self.page.wait_for_selector(loc.CHAT_INPUT_BOX, timeout=timeout)
            return True

        except PlaywrightTimeoutError:
            print(f"❌ Timeout esperando el input del chat '{chat_name}'")
            return False
        except Exception as e:
            print(f"❌ Error al abrir el chat '{chat_name}': {e}")
            return False

    async def new_group(self, group_name: str, members: List[str]) -> Optional[ElementHandle]:
        print(f"Creating new group: {group_name} with members: {members}")
        """
        Crea un nuevo grupo con el nombre especificado
        """
        try:
            # Hacer click en el botón de nuevo chat
            new_chat_button = await self.page.wait_for_selector(
                loc.NEW_CHAT_BUTTON, timeout=5000
            )
            if new_chat_button:
                await new_chat_button.click()
            new_group_button = await self.page.wait_for_selector(
                loc.NEW_GROUP_BUTTON, timeout=5000
            )
            if new_group_button:
                await new_group_button.click()
            # Esperar al campo de nombre del grupo
            member_name_input = await self.page.wait_for_selector(
                loc.INPUT_MEMBERS_GROUP, timeout=5000
            )
            if member_name_input:
                for name in members:
                    await member_name_input.fill(name)
                    await asyncio.sleep(0.5)  # Esperar un poco entre entradas
                    await self.page.keyboard.press("Enter")
                    
            enter_arrow = await self.page.wait_for_selector(
                "xpath=//span[@data-icon='arrow-forward']", timeout=5000
            )
            if enter_arrow:
                await enter_arrow.click()
                
            input_group_name = await self.page.wait_for_selector(
                loc.ENTER_GROUP_NAME, timeout=5000
            )
            if input_group_name:
                await input_group_name.fill(group_name)
                await self.page.keyboard.press("Enter")
            


        except PlaywrightTimeoutError:
            print("Timeout while trying to create a new group")
            return None
        except Exception as e:
            print(f"Error creating new group: {e}")
            return None
            
    async def add_members_to_group(
        self, group_name: str, members: List[str]
    ) -> bool:
        """
        Agrega miembros a un grupo existente. Asume que el chat del grupo ya está abierto.
        """
        try:
            if not self.open(group_name, timeout=5000):
                print(f"❌ No se pudo abrir el grupo '{group_name}'")
                return False
            
            # 2. Hacer clic en la cabecera para abrir la info del grupo
            header = await self.page.wait_for_selector(loc.GROUP_INFO_BUTTON, timeout=5000)
            await header.click()

            # 2. Buscar y hacer clic en el botón "Add participant"
            # Usamos un selector de texto porque es más robusto
            add_participant_button = await self.page.wait_for_selector(
                loc.ADD_MEMBERS_BUTTON, timeout=5000
            )
            await add_participant_button.click()

            # 3. Agregar cada miembro
            member_input = await self.page.wait_for_selector(
                loc.INPUT_MEMBERS_GROUP, timeout=5000
            )
            for member in members:
                await member_input.fill(member)
                await asyncio.sleep(0.5)
                await self.page.keyboard.press("Enter")
                await asyncio.sleep(0.5)

            # 4. Confirmar la adición
            confirm_button = await self.page.wait_for_selector(
                loc.CONFIRM_ADD_MEMBERS_BUTTON, timeout=5000
            )
            await confirm_button.click()
            await asyncio.sleep(0.5)  # Esperar un poco para que se procese
            
            confirm_add_button = await self.page.wait_for_selector('//div[text()="Add member"]', timeout=3000)
            
            # Esperar un poco para que se procese y cerrar el panel
            await asyncio.sleep(1)
            await self.page.keyboard.press("Escape")
            return True

        except PlaywrightTimeoutError:
            print(f"Timeout al intentar agregar miembros a '{group_name}'")
            await self.page.keyboard.press("Escape") # Intentar limpiar
            return False
        except Exception as e:
            print(f"Error agregando miembros a '{group_name}': {e}")
            await self.page.keyboard.press("Escape") # Intentar limpiar
            return False
    async def del_member_group(self, group_name: str, member_name: str) -> bool:
        """
        Elimina un miembro de un grupo existente. Asume que el chat del grupo ya está abierto.
        """
        try:
            if not await self.open(group_name, timeout=5000):
                print(f"❌ No se pudo abrir el grupo '{group_name}'")
                return False
            
            # 2. Hacer clic en la cabecera para abrir la info del grupo
            header = await self.page.wait_for_selector(loc.GROUP_INFO_BUTTON, timeout=5000)
            await header.click()

            # 3. Buscar al miembro, activar menú contextual y hacer clic
            xpath_member_row = f'//span[@title="{member_name}"]/ancestor::div[@role="button"]'
            member_row = await self.page.query_selector(f"xpath={xpath_member_row}")

            if member_row:
                await member_row.scroll_into_view_if_needed()
                await member_row.hover()

                # Dale tiempo a que aparezca
                try:
                    context_menu_btn = await member_row.wait_for_selector(
                        'button[aria-label="Open the chat context menu"]',
                        timeout=3000
                    )
                    await context_menu_btn.click()
                    print("✅ Menú contextual clickeado correctamente.")
                except Exception as e:
                    print(f"❌ No se pudo hacer clic en el botón del menú: {e}")
                
                remove_button = await self.page.wait_for_selector(
                    loc.REMOVE_MEMBER_BUTTON, timeout=5000
                )
                await remove_button.click()
                await asyncio.sleep(0.5)
                confirm_button = await self.page.wait_for_selector('//div[text()="Remove"]', timeout=3000)
                await confirm_button.click()
                
                print(f"✅ Miembro '{member_name}' eliminado de '{group_name}'.")
            else:
                print(f"🔍 Miembro '{member_name}' no encontrado en '{group_name}'")
            return False

        except PlaywrightTimeoutError:
            print(f"Timeout al intentar eliminar miembro de '{group_name}'")
            await self.page.keyboard.press("Escape")
            return False
        


"""No authentication strategy implementation"""

from typing import Dict, Any
from .auth import AuthBase
from playwright.async_api import BrowserContext


class NoAuth(AuthBase):
    """Authentication strategy that doesn't persist any data"""

    def __init__(self):
        super().__init__()

    def get_browser_args(self) -> Dict[str, Any]:
        """Get browser launch arguments"""
        return {"headless": False}

    async def setup_context(self, context: BrowserContext) -> None:
        """No special configuration needed"""
        pass

    async def save_session(self) -> bool:
        """No session to save"""
        return False

    async def load_session(self) -> bool:
        """No session to load"""
        return False



import asyncio
import signal
import sys
import time
from typing import Optional, Dict, List, Any, Union

from .base_client import BaseWhatsAppClient
from .wa_elements import WhatsAppElements
from .constants.states import State
from .chat_manager import ChatManager
from .state_manager import StateManager
from .object.message import Message, FileMessage



class Client(BaseWhatsAppClient):
    """
    Cliente de WhatsApp Web implementado con Playwright
    """

    def __init__(
        self,
        user_data_dir: Optional[str] = None,
        headless: bool = False,
        locale: str = "en-US",
        auth: Optional[Any] = None,
    ):
        super().__init__(user_data_dir=user_data_dir, headless=headless, auth=auth)
        self.locale = locale
        self._cached_chats = set()
        self.poll_freq = 0.25
        self.wa_elements = None
        self.qr_task = None
        self.current_state = None
        self.unread_messages_sleep = 1
        self._shutdown_event = asyncio.Event()
        self._consecutive_errors = 0
        self.last_qr_shown = None
        self.chat_manager = None
        self.state_manager = None
        self._setup_signal_handlers()

    def _setup_signal_handlers(self):
        """Configura los manejadores de señales para un cierre limpio"""
        if sys.platform != "win32":
            for sig in (signal.SIGINT, signal.SIGTERM):
                try:
                    asyncio.get_event_loop().add_signal_handler(
                        sig, lambda s=sig: asyncio.create_task(self._handle_signal(s))
                    )
                except (NotImplementedError, RuntimeError):
                    signal.signal(
                        sig, lambda s, f: asyncio.create_task(self._handle_signal(s))
                    )
        else:
            for sig in (signal.SIGINT, signal.SIGTERM):
                signal.signal(
                    sig, lambda s, f: asyncio.create_task(self._handle_signal(s))
                )

    async def _handle_signal(self, signum):
        """Maneja las señales del sistema para un cierre limpio"""
        signame = (
            signal.Signals(signum).name if hasattr(signal, "Signals") else str(signum)
        )
        print(f"\nRecibida señal {signame}. Cerrando limpiamente...")
        self._shutdown_event.set()
        await self.stop()
        sys.exit(0)

    @property
    def running(self) -> bool:
        return getattr(self, "_is_running", False)

    async def stop(self):
        if not hasattr(self, "_is_running") or not self._is_running:
            return

        self._is_running = False
        try:
            if hasattr(self, "_page") and self._page:
                try:
                    await self._page.close()
                except Exception as e:
                    await self.emit("on_error", f"Error al cerrar la página: {e}")
                finally:
                    self._page = None
            await super().stop()
            if hasattr(self, "_browser") and self._browser:
                try:
                    await self._browser.close()
                except Exception as e:
                    await self.emit("on_error", f"Error al cerrar el navegador: {e}")
                finally:
                    self._browser = None
            if hasattr(self, "playwright") and self.playwright:
                try:
                    await self.playwright.stop()
                except Exception as e:
                    await self.emit("on_error", f"Error al detener Playwright: {e}")
                finally:
                    self.playwright = None
        except Exception as e:
            await self.emit("on_error", f"Error durante la limpieza: {e}")
        finally:
            await self.emit("on_stop")
            self._shutdown_event.set()

    async def start(self) -> None:
        try:
            await super().start()
            self.wa_elements = WhatsAppElements(self._page)
            self.chat_manager = ChatManager(self)
            self.state_manager = StateManager(self)
            self._is_running = True
            await self._main_loop()
        except asyncio.CancelledError:
            await self.emit("on_info", "Operación cancelada")
            raise
        except Exception as e:
            await self.emit("on_error", f"Error en el bucle principal: {e}")
            raise
        finally:
            await self.stop()

    async def _main_loop(self) -> None:
        if not self._page:
            await self.emit("on_error", "No se pudo inicializar la página")
            return

        await self.emit("on_start")
        try:
            await self._page.screenshot(path="init_main.png", full_page=True)
        except Exception as e:
            await self.emit("on_warning", f"No se pudo tomar captura inicial: {e}")
        await self._run_main_loop()

    async def _run_main_loop(self) -> None:
        state = None
        while self._is_running and not self._shutdown_event.is_set():
            try:
                curr_state = await self.state_manager._get_state()
                self.current_state = curr_state

                if curr_state is None:
                    await asyncio.sleep(self.poll_freq)
                    continue

                if curr_state != state:
                    await self.state_manager._handle_state_change(curr_state, state)
                    state = curr_state
                else:
                    await self.state_manager._handle_same_state(curr_state)

                await self.emit("on_tick")
                await asyncio.sleep(self.poll_freq)

            except asyncio.CancelledError:
                await self.emit("on_info", "Bucle principal cancelado")
                raise
            except Exception as e:
                await self.emit("on_error", f"Error en la iteración del bucle: {e}")
                await asyncio.sleep(1)
                if self._consecutive_errors > 5:
                    await self.emit(
                        "on_warning",
                        "Demasiados errores consecutivos, intentando reconectar...",
                    )
                    try:
                        await self.reconnect()
                        self._consecutive_errors = 0
                    except Exception as reconnect_error:
                        await self.emit(
                            "on_error", f"Error al reconectar: {reconnect_error}"
                        )
                        break

    async def wait_until_logged_in(self, timeout: int = 60) -> bool:
        start = time.time()
        while time.time() - start < timeout:
            if self.current_state == State.LOGGED_IN:
                return True
            await asyncio.sleep(self.poll_freq)
        await self.emit("on_error", "Tiempo de espera agotado para iniciar sesión")
        return False

    # Delegated methods to ChatManager
    async def close(self):
        return await self.chat_manager.close()

    async def open(self, chat_name: str, timeout: int = 10000, force_open: bool = False) -> bool:
        return await self.chat_manager.open(chat_name, timeout, force_open)

    async def search_conversations(self, query: str, close=True) -> List[Dict[str, Any]]:
        return await self.chat_manager.search_conversations(query, close)

    async def collect_messages(self) -> List[Union[Message, FileMessage]]:
        return await self.chat_manager.collect_messages()

    async def download_all_files(self, carpeta: Optional[str] = None):
        return await self.chat_manager.download_all_files(carpeta)

    async def download_file_by_index(self, index: int, carpeta: Optional[str] = None):
        return await self.chat_manager.download_file_by_index(index, carpeta)

    async def send_message(self, chat_query: str, message: str, force_open=True) -> bool:
        return await self.chat_manager.send_message(chat_query, message, force_open)

    async def send_file(self, chat_name, path):
        return await self.chat_manager.send_file(chat_name, path)

    async def new_group(self, group_name: str, members: list[str]):
        return await self.chat_manager.new_group(group_name, members)

    async def add_members_to_group(self, group_name: str, members: list[str]):
        return await self.chat_manager.add_members_to_group(group_name, members)


"""
Main module exports for whatsplay library
"""

from whatsplay.client import Client
from whatsplay.base_client import BaseWhatsAppClient
from whatsplay.events.event_handler import EventHandler
from whatsplay.auth.local_profile_auth import LocalProfileAuth
from whatsplay.auth.no_auth import NoAuth
from whatsplay.chat_manager import ChatManager
from whatsplay.state_manager import StateManager

__version__ = "0.1.0"

__all__ = ["Client", "BaseWhatsAppClient", "EventHandler", "NoAuth", "LocalProfileAuth", "ChatManager", "StateManager"]



# State locators
AUTH = "//div[contains(text(), 'Steps to log in')]"
QR_CODE = "//canvas[@aria-label='Scan this QR code to link a device!']"
LOADING = "//div[//span[@data-icon='lock-refreshed'] and contains(text(), 'End-to-end encrypted')]"
LOADING_CHATS = "//div[text()='Loading your chats']"
LOGGED_IN = "//span[@data-icon='wa-wordmark-refreshed']"

# Navigation buttons
CHATS_BUTTON = "//div[@aria-label='Chats']"
STATUS_BUTTON = "//div[@aria-label='Status']"
CHANNELS_BUTTON = "//div[@aria-label='Channels']"
COMMUNITIES_BUTTON = "//div[@aria-label='Communities']"

# Chat filter buttons
ALL_CHATS_BUTTON = "//div[text()='All']"
UNREAD_CHATS_BUTTON = "//div[text()='Unread']"
FAVOURITES_CHATS_BUTTON = "//div[text()='Favourites']"
GROUPS_CHATS_BUTTON = "//div[text()='Groups']"

# Search related locators
SEARCH_BUTTON = [
    "//button[@aria-label='Search']",
    "//button[@title='Search']",
    "//button[@aria-label='Search or start new chat']",
    "//div[@role='button' and @title='Search input textbox']",
    "//span[@data-icon='search']/parent::button",
    "//span[@data-testid='search']/parent::button",
]

SEARCH_TEXT_BOX = [
    "//div[@contenteditable='true']",
    "//div[contains(@class, 'lexical-rich-text-input')]//div[@contenteditable='true']",
    "//div[@role='textbox'][@contenteditable='true']",
    "//div[contains(@class, '_13NKt')]",
]

SEARCH_RESULT = "//div[@aria-label='Search results.']"
SEARCH_ITEM = "//div[@role='listitem']"
CHAT_INPUT_BOX = "//div[@title='Type a message'][@role='textbox']"
CANCEL_SEARCH = "//button[@aria-label='Cancel search']"

# Chat interface elements
CHAT_INPUT_BOX = "//div[@aria-placeholder='Type a message']"
CHAT_DIV = "//div[@role='application']"
UNREAD_CHAT_DIV = "//div[@aria-label='Chat list']"

# Search results
SEARCH_RESULT = "//div[@aria-label='Search results.']"
SEARCH_ITEM = "//div[@role='listitem']"
SEARCH_ITEM_COMPONENTS = (
    ".//div[@role='gridcell' and @aria-colindex='2']/parent::div/div"
)
SEARCH_ITEM_UNREAD_MESSAGES = ".//span[contains(@aria-label, 'unread message')]"
SPAN_TITLE = ".//span[@title]"

# Message elements
CHAT_COMPONENT = ".//div[@role='row']"
CHAT_MESSAGE = ".//div[@data-pre-plain-text"
CHAT_MESSAGE_QUOTE = ".//div[@aria-label='Quoted message']"
CHAT_MESSAGE_IMAGE = ".//div[@aria-label='Open picture']"
CHAT_MESSAGE_IMAGE_ELEMENT = (
    ".//img[starts-with(@src, 'blob:https://web.whatsapp.com')]"
)
# Locator para identificar cualquier botón de descarga de archivos
ANY_DOWNLOAD_ICON = "//span[@data-icon='audio-download']"
ATTACH_BUTTON = "span[data-icon='plus-rounded']"
SEND_BUTTON = "span[data-icon='wds-ic-send-filled']"
FILE_INPUT = "input[type='file']"
NEW_CHAT_BUTTON = 'xpath=//span[@data-icon="new-chat-outline"]'
NEW_GROUP_BUTTON = 'xpath=//div[@aria-label="New group" and @role="button"]'
INPUT_MEMBERS_GROUP = 'xpath=//input[@placeholder="Search name or number"]'
ENTER_GROUP_NAME = 'xpath=//div[@aria-label="Group subject (optional)"]'
GROUP_INFO_BUTTON = 'xpath=//div[@title="Profile details" and @role="button"]'
ADD_MEMBERS_BUTTON = 'xpath=//span[@data-icon="person-add-filled-refreshed"]'
CONFIRM_ADD_MEMBERS_BUTTON = 'xpath=//span[@aria-label="Confirm" and @data-icon="checkmark-medium"]'
REMOVE_MEMBER_BUTTON = 'xpath=//span[@data-icon="clear-refreshed"]'